/************** Machine (and compiler) dependent definitions. **************
 *
 *	This file is for the Symmetry.
 *	Use with s-dynix3-0.h.
 */



/*      MACHINE TYPE	DEFINED TYPE		VALUE RANGE	*/

typedef unsigned char	int8;		/*        0 ..     255 */
typedef short		int16;		/*  -10,000 ..  10,000 */
typedef long		int32;		/* -100,000 .. 100,000 */
typedef unsigned long	uint32;		/* 	  0 ..  2^31-1 */

/*
 * The Symmetry C compiler lacks vprintf/vsprintf and strcspn
 */

#define NO_VARARGS
#define STRCSPN

/* enable parallel make hook in xmakefile */
#define PARALLEL_MAKE

/*
 *	Not in network byte order on the 386
 */

#undef NETWORK_BYTE_ORDER	/* */
#ifdef NETWORK_DATABASE
#include <netinet/in.h>
#endif
